#include "sonyserial.h"
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <math.h>

// http://www.belle-nuit.com/archives/9pin.html

//#define DODUMP

// TODO: fix hadcoded buflen (at least evtbuf)
struct _sonyserial_t {
  int fd;
  struct termios old_tio;
  int lastsync; // 15 NO SYNC, 0 HAVE SYNC
  unsigned char buf[20];
  int len;

  unsigned char devtype[2];
  unsigned char status[9];

  bool localenable;
  unsigned char editpreset[2];

  sonyevent_t evtbuf[20];
  int buflen;

  TIMECODE_CB tccb;
};

static void set_mode(int fd,struct termios *old) // {{{
{
  struct termios tio;

  if (old) {
    tcgetattr(fd,old);
  }
  tcgetattr(fd,&tio);

  tio.c_iflag=INPCK;
  tio.c_oflag&=~(OPOST);
  tio.c_cflag=CS8|CREAD|PARENB|PARODD| CLOCAL;
  tio.c_lflag&=~(ICANON|ECHO|ECHOE|ISIG);  // =0;
//  tio.c_cc[VMIN]=0;
//tio.c_iflag&=~INPCK;

  cfsetspeed(&tio,B38400);

  int ret=tcsetattr(fd,TCSAFLUSH,&tio);
  assert(ret==0);
}
// }}}

sonyserial_t *open_sonyserial(const char *device) // {{{
{
  sonyserial_t *ret;

  ret=malloc(sizeof(sonyserial_t));
  if (!ret) {
    return NULL;
  }

  ret->fd=open(device,O_RDWR|O_NOCTTY);
  if (ret->fd==-1) {
    fprintf(stderr,"Device open failed\n");
    free(ret);
    return NULL;
  }

  // set up 38400 8O1
  set_mode(ret->fd,&ret->old_tio);

  ret->lastsync=15;
  ret->len=0;

  memcpy(ret->devtype,"\x31\x10",2); // DVR-2000(PAL) for now
  memcpy(ret->status,"\x00\x20\x80\x00\x00\x00\x00\x00\x00",9); // default status

  ret->localenable=true;
  memcpy(ret->editpreset,"\x00\x00",2);

  ret->buflen=0;
  ret->tccb=0;

  tcflush(ret->fd,TCIFLUSH); // clear input

  return ret;
}
// }}}

void close_sonyserial(sonyserial_t *ser) // {{{
{
  if (ser) {
    tcsetattr(ser->fd,TCSAFLUSH,&ser->old_tio);
    close(ser->fd);
    free(ser);
  }
}
// }}}

void dump_block(unsigned char *buf) // {{{
{
  const unsigned char cmd1=*buf>>4,cnt=*buf&0xf;
  static const char *cmds[16]={"System control","System control RET",
                               "Transport control","Unknown(3)","Preset/Select control","Unknown(5)",
                               "Sense Request","Sense RETurn",
                               "Unknown(8)","Unknown(9)","Unknown(10)","Unknown(11)","Unknown(12)","Unknown(13)","Unknown(14)","Unknown(15)"};
  printf("Cmd1: 0x%02x, %s, length: %d, Cmd2: 0x%02x",buf[0],cmds[cmd1],cnt,buf[1]);
  if (cnt>0) {
    printf("\n");
  }
  unsigned char csum=buf[0]+buf[1];
  int iA;
  for (iA=0;iA<cnt;iA++) {
    printf(" %02x",buf[iA+2]);
    csum+=buf[iA+2];
  }
  printf(" - Checksum: 0x%02x  (got 0x%02x)\n",buf[cnt+2],csum);
}
// }}}

static int get_length(unsigned char *buf) // {{{
{
  return *buf&0xf;
}
// }}}

static int calculate_checksum(unsigned char *buf) // {{{
{
  unsigned char cnt=(*buf&0xf)+2;
  unsigned char csum=0;
  int iA;
  for (iA=0;iA<cnt;iA++) {
    csum+=buf[iA];
  }
  return csum;
}
// }}}

static int check_checksum(unsigned char *buf) // {{{
{
  unsigned char cnt=(*buf&0xf)+2;
  return (calculate_checksum(buf)==buf[cnt]);
}
// }}}

static int send_response(sonyserial_t *ser,unsigned char cmd1,unsigned char cmd2,unsigned char len,unsigned char *data) // {{{
{
  assert(ser);
  assert( (len>=0)&&(len<=15) );
  assert( (len==0)||(data) );

  unsigned char buf[20];
  buf[0]=(cmd1&0xf0)|(len&0x0f);
  buf[1]=cmd2;
  if (len>0) {
    memcpy(buf+2,data,len);
  }
  buf[len+2]=calculate_checksum(buf);
#ifdef DODUMP
  dump_block(buf);
#endif
  int res=write(ser->fd,buf,3+len);
  return res;
}
// }}}

static void send_ack(sonyserial_t *ser) // {{{
{
  int res=send_response(ser,0x10,0x01,0,NULL);
  assert(res>0);
}
// }}}

#define NAK_TIMEOUT 0x80
#define NAK_FRAMING 0x40
#define NAK_OVERRUN 0x20
#define NAK_PARITY  0x10
#define NAK_CHECKSUM 0x04
#define NAK_UNDEFCMD 0x01
static void send_nak(sonyserial_t *ser,unsigned char error) // {{{
{
  int res=send_response(ser,0x10,0x02,1,&error);
  assert(res>0);
}
// }}}


// expects 9 bytes
void sonyserial_setstatus(sonyserial_t *ser,unsigned char *status) // {{{
{
  assert(ser);
  memcpy(ser->status,status,9);
}
// }}}

unsigned char *sonyserial_getstatus(sonyserial_t *ser) // {{{
{
  assert(ser);
  return ser->status;
}
// }}}


static bool add_event(sonyserial_t *ser,sonyevent_t *evt) // {{{
{
  assert(evt);
  if (ser->buflen>18) { // overflow  TODO? better
    memcpy(ser->evtbuf,ser->evtbuf+1,(ser->buflen-1)*sizeof(sonyevent_t));
    ser->evtbuf[ser->buflen]=*evt;
    return false;
  }
  ser->evtbuf[ser->buflen++]=*evt;
  return true;
}
// }}}

static bool handle_sc(sonyserial_t *ser,unsigned char *buf) // {{{
{
  const unsigned char cmd2=buf[1];
  sonyevent_t evt;
  switch (cmd2) {
  case 0x0c: 
    evt.type=SEV_SC_LOCAL;
    evt.enable=false;
    ser->localenable=false;
    break;
  case 0x1d: 
    evt.type=SEV_SC_LOCAL;
    evt.enable=true;
    ser->localenable=true;
    break;
  case 0x11: // device type
    { int res=send_response(ser,0x10,cmd2,2,ser->devtype);
    assert(res>0);
    return true; }
  default:
    return false;
  }
  if (!add_event(ser,&evt)) {
    send_nak(ser,NAK_OVERRUN);
  } else {
    send_ack(ser);
  }
  return true;
}
// }}}

// TODO?! Edit Off  clears any EE mode  [? one mode enum in sonyserial]
// TODO: EE mode needs representation in status
// TODO: select EE needs  edit preset
// FIXME: Edit On  to use sonyserial's edit preset
static bool handle_tc(sonyserial_t *ser,unsigned char *buf) // {{{
{
  const unsigned char cnt=*buf&0xf,cmd2=buf[1];
  sonyevent_t evt;
  switch (cmd2) {
  case 0x00: evt.type=SEV_TC_STOP; break;
  case 0x01: evt.type=SEV_TC_PLAY; break;
  case 0x02: evt.type=SEV_TC_RECORD; break;
  case 0x04: evt.type=SEV_TC_STANDBY; evt.enable=false;
ser->status[1]&=~0x80;
    break;
  case 0x05: evt.type=SEV_TC_STANDBY; evt.enable=true;
ser->status[1]|=0x80;
    break;
  case 0x0f: evt.type=SEV_TC_EJECT; break;
  case 0x10: evt.type=SEV_TC_FF; break;
  case 0x20: evt.type=SEV_TC_REV; break;
  case 0x11: case 0x12: case 0x13:
  case 0x21: case 0x22: case 0x23:
    evt.type=SEV_TC_STOP+cmd2;
    if (cnt==1) {
      evt.tape_speed=pow(10,buf[2]/32.-2);
      if (buf[2]==0) {
        if (cmd2==0x11) {
          evt.type=SEV_TC_JOGPAUSE;
        } else if (cmd2==0x13) {
          evt.type=SEV_TC_SHUTTLEPAUSE;
        }
      }
    } else if (cnt==2) {
      evt.tape_speed=(1.-buf[3]/256.)*pow(10,buf[2]/32.-2)+
                     buf[3]/256.*pow(10,(buf[2]+1)/32.-2);
    } else {
      return false;
    }
    // ? minimum speed issue
    break;
  case 0x30: evt.type=SEV_TC_PREROLL; break;
  case 0x31:
    if (cnt!=4) {
      return false;
    }
    evt.type=SEV_TC_CUETO;
    evt.tc.hour=(buf[5]>>4)*10+(buf[5]&0x0f);
    evt.tc.minute=(buf[4]>>4)*10+(buf[4]&0x0f);
    evt.tc.second=(buf[3]>>4)*10+(buf[3]&0x0f);
    evt.tc.frame=(buf[2]>>4)*10+(buf[2]&0x0f);
    break;
  case 0x60: 
    evt.type=SEV_TC_FULLEE; evt.enable=false;
ser->status[4]=(ser->status[4]&~0xc0);
    break;
  case 0x61:
    evt.type=SEV_TC_FULLEE; evt.enable=true;
ser->status[4]=(ser->status[4]&~0xc0)|0x80;
    break;
  case 0x63:
    evt.type=SEV_TC_SELECTEE; evt.enable=true;
ser->status[4]=(ser->status[4]&~0xc0)|0x40;
    break;
  case 0x64:
    evt.type=SEV_TC_EDIT; evt.enable=false;
ser->status[4]=(ser->status[4]&~0xc0);
    break;
  case 0x65:
    evt.type=SEV_TC_EDIT; evt.enable=true;
//    ser->status[4]&=~0xc0;
    break;
  default:
    return false;
  }
  if (!add_event(ser,&evt)) {
    send_nak(ser,NAK_OVERRUN);
  } else {
    send_ack(ser);
  }
  return true;
}
// }}}

// TODO? disambiguate editpreset with 1/2 data bytes
// FIXME cue channel in edit preset
static bool handle_ps(sonyserial_t *ser,unsigned char *buf) // {{{
{
  const unsigned char cnt=*buf&0xf,cmd2=buf[1];
  sonyevent_t evt;
  switch (cmd2) {
  case 0x30:
    evt.type=SEV_PS_EDITPRESET;
    if (cnt==1) {
      ser->editpreset[0]=buf[2];
    } else if (cnt==2) {
      ser->editpreset[0]=buf[2];
      ser->editpreset[1]=buf[3];
    } else {
      return false;
    }
    memcpy(evt.preset,ser->editpreset,2); // TODO cnt==1?
    break;
  case 0x35: // SEV_PS_COLORFRAME
  case 0x36: // SEV_PS_TIMERMODE
    if (cnt!=1) {
      return false;
    }
    evt.type=(SEV_PS_COLORFRAME&~0xff)|cmd2;
    evt.mode=buf[2];
    break;
  case 0x40: evt.type=SEV_PS_AUTO; evt.enable=false; break;
  case 0x41: evt.type=SEV_PS_AUTO; evt.enable=true; break;
  case 0x42: evt.type=SEV_PS_SPOTERASE; evt.enable=false; break;
  case 0x43: evt.type=SEV_PS_SPOTERASE; evt.enable=true; break;
  case 0x44: evt.type=SEV_PS_AUDIOSPLIT; evt.enable=false; break;
  case 0x45: evt.type=SEV_PS_AUDIOSPLIT; evt.enable=true; break;
  default:
    return false;
  }
  if (!add_event(ser,&evt)) {
    send_nak(ser,NAK_OVERRUN);
  } else {
    send_ack(ser);
  }
  return true;
}
// }}}

static bool handle_sense(sonyserial_t *ser,unsigned char *buf) // {{{
{
  const unsigned char cnt=*buf&0xf,cmd2=buf[1];
  switch (cmd2) {
  case 0x20:  // status sense
    if (cnt!=1) {
      return false;
    }
    const unsigned char start=buf[2]>>4,len=buf[2]&0x0f;
    if (start+len<=9) {
      int res=send_response(ser,0x70,cmd2,len,ser->status+start);
      assert(res>0);
    } else {
      send_nak(ser,NAK_UNDEFCMD);
    }
    break;
  case 0x0c: // current time
    if (cnt!=1) {
      return false;
    }
    if (buf[2]==0x03) { // "best mode"
//      int res=send_response(ser,0x70,0x06,4,(unsigned char[]){0x00,0x00,0x10,0x00}); // VITC
      timecode_t tc={0,0,0,0};
      if (ser->tccb) {
        tc=(*ser->tccb)();
      }
      int res=send_response(ser,0x70,0x06,4,(unsigned char[]){
          (tc.frame/10)*16+tc.frame%10,
          (tc.second/10)*16+tc.second%10,
          (tc.minute/10)*16+tc.minute%10,
          (tc.hour/10)*16+tc.hour%10}); // VITC
      assert(res>0);
    } else if (buf[2]==0x30) { // user bits of "best mode"
      // FIXME
      int res=send_response(ser,0x70,0x07,4,(unsigned char[]){
          0,0,0,0}); // VITC user bits
      assert(res>0);
      fprintf(stderr,"FIXME: TC user bits\n");
    } else {
      return false;
    }
    break;
  default:
    return false;
  }
  return true;
}
// }}}

// assumes checked checksum
static void process_block(sonyserial_t *ser,unsigned char *buf) // {{{
{
  const unsigned char cmd1=*buf>>4;

#ifdef DODUMP
  dump_block(buf);
#endif
  switch (cmd1) {
  case 0: // system control
    if (!handle_sc(ser,buf)) {
      fprintf(stderr,"Bad/Unknown system control:\n");
#ifndef DODUMP
      dump_block(buf);
#endif
      send_nak(ser,NAK_UNDEFCMD);
    }
    break;
  case 2: // transport control
    if (!handle_tc(ser,buf)) {
      fprintf(stderr,"Bad/Unknown transport control:\n");
#ifndef DODUMP
      dump_block(buf);
#endif
      send_nak(ser,NAK_UNDEFCMD);
    }
    break;
  case 4: // preset/select control
    if (!handle_ps(ser,buf)) {
      fprintf(stderr,"Bad/Unknown preset/select control:\n");
#ifndef DODUMP
      dump_block(buf);
#endif
      send_nak(ser,NAK_UNDEFCMD);
    }
    break;
  case 6: // sense request
    if (!handle_sense(ser,buf)) {
      fprintf(stderr,"Bad/Unknown sense request:\n");
#ifndef DODUMP
      dump_block(buf);
#endif
      send_nak(ser,NAK_UNDEFCMD);
    }
    break;
  default:
    fprintf(stderr,"Unrecognized command:\n");
#ifndef DODUMP
    dump_block(buf);
#endif
    send_nak(ser,NAK_UNDEFCMD);
  }
}
// }}}

// -1: none, -2: error, otherwise the read character
int sonyserial_readone(sonyserial_t *ser) // {{{
{
  assert(ser);
  unsigned char ret;
  int len=read(ser->fd,&ret,1);
  if (len==-1) {
    return -2;
  } else if (len==0) {
    return -1;
  }
  if (ser->len>=18) { // drop data
    send_nak(ser,NAK_OVERRUN);
    ser->len=0;
  }
  ser->buf[ser->len]=ret;
  ser->len++;

  // update sync
  if (ser->lastsync<15) {
    ser->lastsync++;
  }
  const int slen=get_length(ser->buf);
  if (slen+3==ser->len) {
    if (check_checksum(ser->buf)) {
      process_block(ser,ser->buf);
      ser->len=0;
      ser->lastsync=0;
    }
  } else if ( (slen+3>ser->len)&&(ser->lastsync==1) ) {
    ser->lastsync=0;
  } else if (ser->lastsync==1) {
    send_nak(ser,NAK_CHECKSUM);
  }
  if (ser->lastsync>0) { // not yet done
    int len=ser->len-3;
    if (len>ser->lastsync) {
      len=ser->lastsync;
    }
    int iA;
    for (iA=1;iA<len;iA++) {
      const int slen=get_length(ser->buf+iA);
      if (slen+3<=ser->len-iA) {
        if (check_checksum(ser->buf+iA)) {
          fprintf(stderr,"Skipped %d bytes\n",iA);
          process_block(ser,ser->buf+iA);
          memcpy(ser->buf,ser->buf+iA+slen,ser->len-slen-3);
          ser->len-=slen+3;
          ser->lastsync=0;
          break;
        }
      }
    }
  }

  return ret;
}
// }}}

void dump_event(sonyevent_t *evt) // {{{
{
  if (!evt) {
    printf("(null)\n");
    return;
  }
  printf("Type: ");
  switch (evt->type) {
  case SEV_NONE: printf("SEV_NONE"); break;

  case SEV_SC_LOCAL: printf("SEV_SC_LOCAL - Enable: %d",evt->enable); break;

  case SEV_TC_STOP: printf("SEV_TC_STOP"); break;
  case SEV_TC_PLAY: printf("SEV_TC_PLAY"); break;
  case SEV_TC_RECORD: printf("SEV_TC_RECORD"); break;
  case SEV_TC_STANDBY: printf("SEV_TC_STANDBY - On: %d",evt->enable); break;
  case SEV_TC_EJECT: printf("SEV_TC_EJECT"); break;
  case SEV_TC_JOGPAUSE: printf("SEV_TC_JOGPAUSE"); break;
  case SEV_TC_SHUTTLEPAUSE: printf("SEV_TC_SHUTTLEPAUSE"); break;
  case SEV_TC_FF: printf("SEV_TC_FF"); break;
  case SEV_TC_REV: printf("SEV_TC_REV"); break;
  case SEV_TC_JOGFF: printf("SEV_TC_JOGFF - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_VARFF: printf("SEV_TC_VARFF - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_SHUTTLEFF: printf("SEV_TC_SHUTTLEFF - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_JOGREV: printf("SEV_TC_JOGREV - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_VARREV: printf("SEV_TC_VARREV - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_SHUTTLEREV: printf("SEV_TC_SHUTTLEREV - Speed: %fx",evt->tape_speed); break;
  case SEV_TC_PREROLL: printf("SEV_TC_PREROLL"); break;
  case SEV_TC_CUETO: printf("SEV_TC_CUETO - Timecode: %d.%d.%d.%d",
    evt->tc.hour,evt->tc.minute,evt->tc.second,evt->tc.frame); break;
  case SEV_TC_FULLEE: printf("SEV_TC_FULLEE - On: %d",evt->enable); break;
  case SEV_TC_SELECTEE: printf("SEV_TC_SELECTEE - On: %d",evt->enable); break;
  case SEV_TC_EDIT: printf("SEV_TC_EDIT - On: %d",evt->enable); break;

  case SEV_PS_EDITPRESET: printf("SEV_PS_EDITPRESET - 0x%02x, 0x%02x",evt->preset[0],evt->preset[1]); break;
  case SEV_PS_COLORFRAME: printf("SEV_PS_COLORFRAME - Mode: 0x%02x",evt->mode); break;
  case SEV_PS_TIMERMODE: printf("SEV_PS_TIMERMODE - Mode: 0x%02x",evt->mode); break;
  case SEV_PS_AUTO: printf("SEV_PS_AUTO - On: %d",evt->enable); break;
  case SEV_PS_SPOTERASE: printf("SEV_PS_SPOTERASE - On: %d",evt->enable); break;
  case SEV_PS_AUDIOSPLIT: printf("SEV_PS_AUDIOSPLIT - On: %d",evt->enable); break;

  default:
    printf("0x%02x (?)",evt->type);
  }
  printf("\n");
}
// }}}

bool sonyserial_hasevent(sonyserial_t *ser) // {{{
{
  assert(ser);
  return (ser->buflen>0);
}
// }}} 

sonyevent_t sonyserial_getevent(sonyserial_t *ser) // {{{
{
  assert(ser);

  if (ser->buflen==0) {
    sonyevent_t ret;
    ret.type=SEV_NONE;
    return ret;
  }

  sonyevent_t ret=ser->evtbuf[0];
  ser->buflen--;
  memcpy(ser->evtbuf,ser->evtbuf+1,ser->buflen*sizeof(sonyevent_t));
  return ret;
}
// }}}

void sonyserial_settimecode_cb(sonyserial_t *ser,TIMECODE_CB cb) // {{{
{
  assert(ser);
  ser->tccb=cb;
}
// }}}

